# -*- coding: utf-8; -*-
"""
CORE-POS model importers

.. warning::
   As of this writing, most classes in this module are "direct DB" importers,
   which will write directly to MySQL.  They are meant to be used in dry-run
   mode only, and/or for sample data import to a dev system etc.  They are
   *NOT* meant for production use, as they will completely bypass any CORE
   business rules logic which may exist.

   However some of the importers will be refactored, so they directly use the
   CORE API for *writing* data (although they may continue to read directly
   from the DB).  The hope is that *all* existing importers can be refactored
   to write via API instead of DB, so for now all importers are left intact
   until they can be dealt with (hopefully refactored, but if not, then
   removed).
"""

from sqlalchemy.orm.exc import NoResultFound

from rattail import importing

from corepos.db.office_op import model as corepos
from corepos.db.office_trans import model as coretrans
from corepos.api import CoreWebAPI


class ToCore(importing.ToSQLAlchemy):
    """
    Base class for all CORE "operational" model importers.

    Note that this class inherits from
    :class:`~rattail:rattail.importing.sqlalchemy.ToSQLAlchemy` even though our
    goal is to *not* write directly to the CORE DB.  However, (for now)
    importers will need to override methods to ensure API is used instead,
    where applicable.  But even once all have been refactored to write via API,
    we *still* may want to keep using ``ToSQLAlchemy`` for the sake of reading
    "cached local" data.  (May depend on how robust the API is.)
    """
    # TODO: should we standardize on the 'id' primary key? (can we even?)
    # key = 'id'

    def setup(self):
        self.establish_api()

    def establish_api(self):
        url = self.config.require('corepos.api', 'url')
        self.api = CoreWebAPI(url)

    # TODO: this looks an awful lot like it belongs in rattail proper
    def get_single_local_object(self, key):
        """
        Fetch a particular record from CORE, via SQLAlchemy.  This is used by
        the Rattail -> CORE datasync consumer.
        """
        query = self.session.query(self.model_class)
        for i, field in enumerate(self.key):
            query = query.filter(getattr(self.model_class, field) == key[i])

        for option in self.cache_query_options():
            query = query.options(option)

        try:
            return query.one()
        except NoResultFound:
            pass


class ToCoreTrans(importing.ToSQLAlchemy):
    """
    Base class for all CORE "transaction" model importers
    """


########################################
# CORE Operational
########################################

class DepartmentImporter(ToCore):
    model_class = corepos.Department
    key = 'number'


class SubdepartmentImporter(ToCore):
    model_class = corepos.Subdepartment
    key = 'number'


class VendorImporter(ToCore):
    model_class = corepos.Vendor
    key = 'vendorID'
    supported_fields = [
        'vendorID',
        'vendorName',
        'vendorAbbreviation',
        'discountRate',
    ]
    # TODO: this importer is in a bit of an experimental state at the moment.
    # we only allow "update" b/c it will use the API instead of direct DB
    allow_create = False
    allow_delete = False

    def normalize_local_object(self, vendor):
        return {
            'vendorID': str(vendor.id),
            'vendorName': vendor.name,
            'vendorAbbreviation': vendor.abbreviation,
            'discountRate': vendor.discount_rate,
        }

    def update_object(self, vendor, data, local_data=None):
        """
        Push an update for the vendor, via the CORE API.
        """
        if self.dry_run:
            return data

        vendorID = data.pop('vendorID')
        self.api.set_vendor(vendorID, **data)
        return data


class VendorContactImporter(ToCore):
    model_class = corepos.VendorContact
    key = 'vendor_id'


class ProductImporter(ToCore):
    model_class = corepos.Product
    key = 'id'


class ProductFlagImporter(ToCore):
    model_class = corepos.ProductFlag
    key = 'bit_number'


class EmployeeImporter(ToCore):
    model_class = corepos.Employee
    key = 'number'


class CustomerImporter(ToCore):
    model_class = corepos.Customer
    key = 'id'


class MemberTypeImporter(ToCore):
    model_class = corepos.MemberType
    key = 'id'


class MemberInfoImporter(ToCore):
    model_class = corepos.MemberInfo
    key = 'card_number'


class MemberDateImporter(ToCore):
    model_class = corepos.MemberDate
    key = 'card_number'


class MemberContactImporter(ToCore):
    model_class = corepos.MemberContact
    key = 'card_number'


class HouseCouponImporter(ToCore):
    model_class = corepos.HouseCoupon
    key = 'coupon_id'


########################################
# CORE Transactions
########################################

class TransactionDetailImporter(ToCoreTrans):
    """
    CORE-POS transaction data importer.
    """
    model_class = coretrans.TransactionDetail
